
#define H_IDENTITY() {0, 0, 0, 0, 0, 0}
#define H_SAFE_MALLOC(s) (s > 0 ? malloc(s) : NULL)
#define H_SAFE_FREE(p) if (p) { free(p); p = NULL; }


struct h_shape
{
	float4 color;
	float2 *point;
	int point_count;
};

struct h_frame
{
	struct h_shape *shape;
	int shape_count;
	int id;
};

struct h_layer
{
	struct h_frame *frame;
	int frame_count;
};

struct h_scene
{
	struct h_layer *layer;
	int layer_count;
	int width;
	int height;
	int fps;
	int duration;
};


void h_shape_destroy(struct h_shape *shape)
{
	H_SAFE_FREE(shape->point);
	shape->point_count = 0;
}

void h_frame_destroy(struct h_frame *frame)
{
	int k;
	
	for (k = 0; k < frame->shape_count; k++)
		h_shape_destroy(frame->shape + k);
	
	frame->shape_count = 0;
	H_SAFE_FREE(frame->shape);
}

void h_destroy(struct h_scene *scene)
{
	int i, j;

	/* layers */
	for (i = 0; i < scene->layer_count; i++) {
		struct h_layer *layer = scene->layer + i;
		/* frames */
		for (j = 0; j < layer->frame_count; j++)
			h_frame_destroy(layer->frame + j);
			
		H_SAFE_FREE(layer->frame);
	}
	
	scene->layer_count = 0;
	H_SAFE_FREE(scene->layer);
}

int h_load(struct h_scene *scene, const char *filename)
{
	FILE *file;
	int version;
	int i, j, k, l;
	
	file = fopen(filename, "rb");
	if (! file) {
		printf("ERROR: unable to open %s\n", filename);
		return 0;
	}

	h_destroy(scene);

	// version
	fread(&version, sizeof(int), 1, file);

	// properties
	fread(&scene->width, sizeof(int), 1, file);
	fread(&scene->height, sizeof(int), 1, file);
	fread(&scene->fps, sizeof(int), 1, file);
	fread(&scene->duration,	sizeof(int), 1, file);

	// layers
	fread(&scene->layer_count, sizeof(int), 1, file);
	scene->layer = H_SAFE_MALLOC(sizeof(struct h_layer) * scene->layer_count);
	for (i = 0; i < scene->layer_count; i++) {
	
		struct h_layer *layer = scene->layer + i;

		// frames
		fread(&layer->frame_count, sizeof(int), 1, file);
		layer->frame = H_SAFE_MALLOC(sizeof(struct h_frame) * layer->frame_count);
		for (j = 0; j < layer->frame_count; j++) {
		
			struct h_frame *frame = layer->frame + j;

			// frame id
			fread(&frame->id, sizeof(int), 1, file);

			// shapes
			fread(&frame->shape_count, sizeof(int), 1, file);
			frame->shape = H_SAFE_MALLOC(sizeof(struct h_shape) * frame->shape_count);

			for (k = 0; k < frame->shape_count; k++) {

				struct h_shape *shape = frame->shape + k;

				// points
				fread(&shape->point_count, sizeof(int), 1, file);
				shape->point = H_SAFE_MALLOC(sizeof(float2) * shape->point_count);

				if (shape->point_count > 0) {

					// color
					fread(&shape->color, sizeof(float), 4, file);
					m_color_sRGB_to_linear((float *)&shape->color,(float *)&shape->color, 4);

					// point data
					fread(shape->point, sizeof(float2), shape->point_count, file);
				}
			}
		}
	}

	fclose(file);
	return 1;
}

void h_shape_translate(struct h_shape *shape, float2 *transl)
{
	int i;
	for (i = 0; i < shape->point_count; i++) {
		shape->point[i].x += transl->x;
		shape->point[i].y += transl->y;
	}
}

void h_shape_scale(struct h_shape *shape, float2 *scale)
{
	int i;
	for (i = 0; i < shape->point_count; i++) {
		shape->point[i].x = shape->point[i].x * scale->x + 0.5f;
		shape->point[i].y = shape->point[i].y * scale->y - 0.5f;
	}
}

void h_shape_draw(struct m_image *dest, const struct h_shape *shape)
{
	float2 scale = {dest->width - 1, dest->height - 1};
	int i;
	
	if (shape->point_count == 0)
		return;
	/*
	for (i = 1; i < shape->point_count; i++) {
	
		float2 p1 = shape->point[i-1];
		float2 p2 = shape->point[i];
		
		m_raster_line(
			(float *)dest->data, dest->width, dest->height, dest->comp,
			(float *)&p1, (float *)&p2,
			(float *)&shape->color);
	}*/

	m_raster_polygon(
		(float *)dest->data, dest->width, dest->height, dest->comp,
		(float *)shape->point, shape->point_count,
		(float *)&shape->color);
}

void h_frame_translate(struct h_frame *frame, float2 *transl)
{
	int k;
	for (k = 0; k < frame->shape_count; k++)
		h_shape_translate(frame->shape + k, transl);
}

void h_frame_scale(struct h_frame *frame, float2 *scale)
{
	int k;
	for (k = 0; k < frame->shape_count; k++)
		h_shape_scale(frame->shape + k, scale);
}

void h_frame_draw(struct m_image *dest, struct h_frame *frame)
{
	int k;
	for (k = 0; k < frame->shape_count; k++)
		h_shape_draw(dest, frame->shape + k);
}

void h_translate(struct h_scene *scene, float2 *transl)
{
	int i, k;
	for (i = 0; i < scene->layer_count; i++) {
		struct h_layer *layer = scene->layer + i;
		for (k = 0; k < layer->frame_count; k++)
			h_frame_translate(layer->frame + k, transl);
	}
}

void h_scale(struct h_scene *scene, float2 *scale)
{
	int i, k;
	for (i = 0; i < scene->layer_count; i++) {
		struct h_layer *layer = scene->layer + i;
		for (k = 0; k < layer->frame_count; k++)
			h_frame_scale(layer->frame + k, scale);
	}
}
